#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/in.h>
#include <linux/irqreturn.h>
#include <linux/proc_fs.h>
#include <net/sock.h>
#include <xen/hypercall.h>
#include <xen/evtchn.h>
#include <xen/interface/xen.h>
#include "ksocket.h"
#include "arpa_inet_lib.h"
#include "kernelsync-client.h"
#include "domctl-mini.h"
#include "linux/time.h"
#include "linux/ktime.h"

MODULE_DESCRIPTION("My kernel module");
MODULE_AUTHOR("Florian Schmidt,,, (flosch@xen-host)");
MODULE_LICENSE("GPL");

/*verbosity settings*/
#define DEBUGLEVEL 1
#define PRINT(_f, _a...)                        \
    do {                                        \
        if ( (_f) <= DEBUGLEVEL )                \
            printk(_a );                        \
    } while ( 0 )


int irq;
static int irqcounter;
struct work_struct bottomhalf;
struct work_struct sync_run;
DECLARE_WORK(bottomhalf, client_irq_bottomhalf, NULL);
DECLARE_WORK(sync_run, bh_run_synced_domain, (void *)&irqcounter);
static DECLARE_COMPLETION(on_exit);

static int server_port = 34512;
module_param(server_port, int, 0);
static int client_port = 34513;
module_param(client_port, int, 0);
static char broadcast_address[16];
module_param_string(broadcast_address, broadcast_address, sizeof(broadcast_address), 0);
static char server_address[16];
module_param_string(server_address, server_address, sizeof(server_address), 0);
static char *cli_descr = "Toller Client";
module_param(cli_descr, charp, 0);
static int barrier_time = 1;
module_param(barrier_time, int, 0);
static int cli_id = 23;
module_param(cli_id, int, 0);
static int sync_domain[MAX_DOMS];
static int sync_domain_count;
module_param_array(sync_domain, int, &sync_domain_count, 0);

static client_context_t clictx;
static int thread_pid;
static rwlock_t clilock;

static sync_packet_t *packet;
static data_reg_t *reg;
static data_unreg_t *unreg;
static data_run_t *run;
static data_fin_t *fin;
static int max_packet_size;

struct timeval before, after;

static inline uint32_t time_as_long(struct timeval af, struct timeval bf)
{
	struct timeval tot;
	tot.tv_sec = af.tv_sec - bf.tv_sec;
	tot.tv_usec = af.tv_usec - bf.tv_usec;
        if (tot.tv_usec < 0) {					      
      		--tot.tv_sec;						     
		tot.tv_usec += 1000000;					      
	}	
	return (uint32_t)tot.tv_sec*1000000 + (uint32_t)tot.tv_usec;
}

static inline void init_rwlock(rwlock_t *lock)
{
#ifdef USE_RWLOCK
  rwlock_init(lock);
#endif
}

static inline void set_readlock(rwlock_t *lock)
{
#ifdef USE_RWLOCK
  read_lock(lock);
#endif
}

static inline void unset_readlock(rwlock_t *lock)
{
#ifdef USE_RWLOCK
  read_unlock(lock);
#endif
}

static inline void set_writelock(rwlock_t *lock)
{
#ifdef USE_RWLOCK
  write_lock(lock);
#endif
}

static inline void unset_writelock(rwlock_t *lock)
{
#ifdef USE_RWLOCK
  write_unlock(lock);
#endif
}

module_init(kernelsync_init_module);
module_exit(kernelsync_exit_module);

static int kernelsync_init_module(void)
{
  int ret;

  printk( KERN_DEBUG "Module kernelsync init\n" );
  
  // maximal size of data part of packet is generally defined byh size of data_reg_t...
  if (CLIENT_MAX_DESCRIPTION_LENGTH > 11)
    max_packet_size = sizeof(data_reg_t);
  // unless the description string field is unusually small
  else
    max_packet_size = sizeof(data_fin_t);
  // allocate a memory area for assembling, sending, and receiving packets.
  packet = kmalloc(sizeof(sync_packet_t)+max_packet_size, GFP_KERNEL);
  // the other pointers all point to the first byte after the header, so we never
  // need to memcpy() anything and can just reuse this strip of memory
  reg = (data_reg_t*)( (unsigned long)&(packet->data));
  unreg = (data_unreg_t*)( (unsigned long)&(packet->data));
  run = (data_run_t*)( (unsigned long)&(packet->data));
  fin = (data_fin_t*)( (unsigned long)&(packet->data));

    printk(KERN_DEBUG "Running in client mode\n");
    init_rwlock(&clilock);
    ret = sync_client_init();
    if (ret < 0)
      return -EIO;
    thread_pid = kernel_thread(client_thread, NULL, CLONE_KERNEL);
    if (thread_pid < 0)
      goto release_sock;
    irq = bind_virq_to_irqhandler(VIRQ_SYNC_DONE, 0, client_irqhandler, SA_INTERRUPT, "syncint", NULL);
    if (irq < 0)
      goto release_sock;
   
  return ret;

release_sock:
  sock_release(clictx.sk);
  return -EIO;
}

static void kernelsync_exit_module(void)
{
  int i, retval;

  printk( KERN_DEBUG "Module kernelsync exit\n" );
  
  // setting destination to self, to send one last dummy packet
  // so thread can exit properly and won't wait forever
  set_writelock(&clilock);
  packet->packet_type = PACKETTYPE_DUMMY;
  clictx.servaddr.sin_addr.s_addr = inet_addr("127.0.0.1");
  clictx.servaddr.sin_port = htons(client_port);
  ksocket_send(clictx.sk, &clictx.servaddr, (unsigned char*)packet, sizeof(sync_packet_t));
  unset_writelock(&clilock);

  // Wait for the thread to complete
  wait_for_completion(&on_exit);

  printk( KERN_DEBUG "irqcounter is %d\n", irqcounter2 );

  // Setting properties back to server stuff so that we can send the unregister message
  // At this point, there is no other thread running, so no need to use locks here
  clictx.servaddr.sin_addr.s_addr = inet_addr(server_address);
  clictx.servaddr.sin_port = htons(server_port);
 
  packet->packet_type = PACKETTYPE_UNREGISTER;
  unreg->id = ntohs(clictx.id);
  unreg->reason = UNREGISTER_REASON_REGULAR;
  retval = send_toserver(packet, sizeof(sync_packet_t)+sizeof(data_unreg_t));
   
  // Release the socket
  sock_release(clictx.sk);
  if (irq >= 0)
    unbind_from_irqhandler(irq, NULL);

  kfree(packet);
  printk(KERN_DEBUG "Done\n");
}

static int sync_client_init(void)
{
  int retval,i;

  set_writelock(&clilock);
  clictx.id = cli_id;
  clictx.period = -1;
  strncpy(cli_descr, clictx.description, CLIENT_MAX_DESCRIPTION_LENGTH);
  
  retval = sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &clictx.sk);
  if (retval < 0)
  {
    printk(KERN_ERR "Error creating socket, error number %d\n", retval);
    return -EIO;
  }
  
  // perusing clictx.servaddr for setting up the client for the time being, correct values are assigned later
  clictx.servaddr.sin_family = AF_INET;
  clictx.servaddr.sin_addr.s_addr = INADDR_ANY;
  clictx.servaddr.sin_port = htons(client_port);

  clictx.sk->sk->sk_reuse = 1; // SO_REUSEADDR
  
  retval = clictx.sk->ops->bind(clictx.sk, (struct sockaddr *) &clictx.servaddr, sizeof(clictx.servaddr));
  if (retval)
  {
    printk(KERN_ERR "Error binding to server socket, error number %d.\n", retval);
    goto release;
  }

  clictx.servaddr.sin_family = AF_INET;
  clictx.servaddr.sin_addr.s_addr = inet_addr(server_address);
  clictx.servaddr.sin_port = htons(server_port);
  
  reg->id = htons(clictx.id);
  unset_writelock(&clilock);
  reg->type = CLIENT_TYPE_TEST;
  snprintf(reg->description, CLIENT_MAX_DESCRIPTION_LENGTH-1, "Kernel UDP foo");
  packet->packet_type = PACKETTYPE_REGISTER;
  
  for (i=0; i<sync_domain_count; i++)
  {
    retval = hypercall_sync_domain(i);
    if (retval)
    {
      printk(KERN_ERR "Error setting sync bits for domain %d, XEN_DOMCTL_set_synced hypercall returned %d\n", sync_domain[i], retval);
      goto release;
    }
  }
  
  // send registering information
  retval = send_toserver(packet, sizeof(sync_packet_t)+sizeof(data_reg_t));
  if (retval)
  {
    printk(KERN_ERR "Error sending register message, ksocket_send returned %d (=bytes sent).\n", retval);
    goto release;
  }

  return 0;
  
release:
  sock_release(clictx.sk);
  return -EIO;

}

static int hypercall_sync_domain(int domnr)
{
  xen_domctl_t dom;
  int ret;
 
  dom.domain = sync_domain[domnr];
  dom.interface_version = XEN_DOMCTL_INTERFACE_VERSION;
  dom.u.synced.is_synchronized = 1;
  dom.cmd = XEN_DOMCTL_set_synced;
  ret = _hypercall1(int, domctl, (unsigned long)&dom);
  printk(KERN_DEBUG "XEN_DOMCTL_set_synced hypercall returned %d\n", ret);
  if (ret)
  {
    return ret;
  }

  return 0;
}

static int hypercall_unsync_domain(int domnr)
{
  xen_domctl_t dom;
  int ret;
  
  dom.domain = sync_domain[domnr];
  dom.interface_version = XEN_DOMCTL_INTERFACE_VERSION;
  dom.u.synced.is_synchronized = 0;
  dom.cmd = XEN_DOMCTL_set_synced;
  ret = _hypercall1(int, domctl, (unsigned long)&dom);
  printk(KERN_DEBUG "Unsync XEN_DOMCTL_set_synced on domain %d hypercall returned %d\n", domnr, ret);
  if (ret)
  {
    return ret;
  }
  return 0;
}
static void bh_run_synced_domain(void *arg)
{ 
  hypercall_run_synced_domain(*(int*)arg);
}

static int hypercall_run_synced_domain(int domnr)
{
  int retval;
  xen_domctl_t dom;

  dom.domain = sync_domain[domnr];
  dom.interface_version = XEN_DOMCTL_INTERFACE_VERSION;
  dom.cmd = XEN_DOMCTL_scheduler_op;
  dom.u.scheduler_op.sched_id = XEN_SCHEDULER_SYNCSEDF;
  dom.u.scheduler_op.cmd = XEN_DOMCTL_SCHEDOP_putinfo;
  dom.u.scheduler_op.u.sedf.latency = 0;
  dom.u.scheduler_op.u.sedf.extratime = 0;
  dom.u.scheduler_op.u.sedf.weight = 0;
  // synchronizer sends the slice size in usecs
  dom.u.scheduler_op.u.sedf.slice = (ntohl(run->runtime)) * 1000;
  // make period double the slice time. this is just a time picked from a few
  // test runs, more of an educated guess
  dom.u.scheduler_op.u.sedf.period = (ntohl(run->runtime))*2000;
  
  
  retval = _hypercall1(int, domctl, (unsigned long)&dom);
  printk(KERN_DEBUG "XEN_DOMCTL_SCHEDOP_putinfo hypercall returned %d\n", retval);
  return retval;
}


int wait_for_runpermission(void)
{
  int retval;
  int period;
  struct sockaddr_in dummy;

  set_current_state(TASK_INTERRUPTIBLE);

  while (1)
  {
    retval = ksocket_receive(clictx.sk, &dummy, (unsigned char*) packet, max_packet_size);

    if (packet->packet_type == PACKETTYPE_RUNPERMISSION)
    {
      period = ntohl(run->period);
      if (unlikely(clictx.period == -1))
      {
	printk(KERN_DEBUG "Initial synchronization of client to server-provided period %u\n", period);
	clictx.period = period;
	return 0;
      }
      // no, this case does not cover the above special case, because this will not be executed
      // if we get a duplicate of a runpermission packet at some point
      if (period > clictx.period) 
      {
	clictx.period = period;
	return 0;
      }
    }
    else if (packet->packet_type == PACKETTYPE_DUMMY)
        return PACKETTYPE_DUMMY;
    else
	{
      		printk(KERN_DEBUG "Packet received, but not a run permission or Dummy packet. Probably Xen bug. Returning Dummy Packet.\n");
		return PACKETTYPE_DUMMY;
	}
  }
  return 1; // should never be reached, just here to make the compiler warning go away
}

int send_toserver(sync_packet_t *packet, int len)
{
  int retval;
  packet->seqnr = htonl(++clictx.seqnr);
  
  printk(KERN_DEBUG "Before ksocket_send, target address 0x%x, port %d, size of packet %d\n", ntohl(clictx.servaddr.sin_addr.s_addr), ntohs(clictx.servaddr.sin_port), len);
  retval = ksocket_send(clictx.sk, &clictx.servaddr, (unsigned char*)packet, len);
  printk(KERN_DEBUG "ksocket_send returned %d\n", retval);
  if (retval == len)
    return 0;
  else
    return retval;
}

static irqreturn_t client_irqhandler(int irq, void *dev_id, struct pt_regs *regs)
{
 printk(KERN_DEBUG "irq received\n");

  irqcounter++;

  if (irqcounter < sync_domain_count)
  { 
    // run next synced domain
    schedule_work(&sync_run);
    goto out;
  }
  irqcounter = 0;
  schedule_work(&bottomhalf);
  
out:
  return IRQ_HANDLED;
}

static void client_irq_bottomhalf(void *arg)
{
  printk(KERN_DEBUG "This is my bottom half. There are many like, but this one is mine.\n");
  do_gettimeofday(&after);  
  set_writelock(&clilock);
  packet->packet_type = PACKETTYPE_FINISHED;
  fin->id = ntohs(clictx.id);
  fin->realtime = htonl(time_as_long(after, before));
  send_toserver(packet, sizeof(sync_packet_t)+sizeof(data_fin_t));

  unset_writelock(&clilock);
}

static int client_thread(void *arg)
{
  int retval,i;
  
  while(1)
  {
    printk(KERN_DEBUG "Waiting for run permission for period %d.\n", clictx.period+1);
    retval = wait_for_runpermission();
    if (retval == PACKETTYPE_DUMMY)
	break;
    printk(KERN_DEBUG "received packet with a runtime of %u\n", ntohl(run->runtime));
    
    set_current_state(TASK_INTERRUPTIBLE);

    do_gettimeofday(&before);
    hypercall_run_synced_domain(0);
   
    set_current_state(TASK_INTERRUPTIBLE);
  }
  
    
  for (i=0; i<sync_domain_count; i++)
    hypercall_unsync_domain(i);

  //Exit without sending any unregistration packet. The main thread will take care of it.
  printk(KERN_DEBUG "This is the client thread, and this line denotes the end. Exiting now.\n");
  complete_and_exit(&on_exit, 0);  
  return 0;
}



